﻿/*
   This is a .NET library that handles calling CryptoPhoto API.

   Be sure toy get your "PrivateKey" and "PublicKey" from:
   https://cryptophoto.com/admin/account

     - Main Page
         http://cryptophoto.com/
     - About Cryptophoto
         http://cryptophoto.com/about
     - Register to CryptoPhoto
         http://cryptophoto.com/demo/register/
 
  Copyright(c) 2014 CryptoPhoto -- http://cryptophoto.com/
  AUTHORS: CryptoPhoto
 
  VERSION: 1.20140728
 
  Permission is hereby granted, free of charge, to any person obtaining a copy
  of this software and associated documentation files (the "Software"), to deal
  in the Software without restriction, including without limitation the rights
  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
  copies of the Software, and to permit persons to whom the Software is
  furnished to do so, subject to the following conditions:
 
  The above copyright notice and this permission notice shall be included in
  all copies or substantial portions of the Software.
 
  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
  THE SOFTWARE.
*/

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Collections.Specialized;
using System.Net;
using System.Security.Cryptography;

namespace CryptoPhoto
{
    public class CryptoPhotoUtils
    {
        private static readonly DateTimeOffset UnixEpoch = new DateTimeOffset(1970, 1, 1, 0, 0, 0, TimeSpan.Zero);

        public const string START_SESSION_URL = "https://cryptophoto.com/api/get/session";
        public const string VERIFY_URL = "https://cryptophoto.com/api/verify";
        public const string GENERATE_WIDGET_URL_COMPOSITE = "https://cryptophoto.com/api/token?sd={0}";
        public const string CHALLENGE_URL_COMPOSITE = "https://cryptophoto.com/api/challenge?sd={0}";

        WebClient webClient;

        public CryptoPhotoUtils(string privateKey, string publicKey)
        {
            this.PrivateKey = privateKey;
            this.PublicKey = publicKey;

            System.Net.ServicePointManager.Expect100Continue = false;
            webClient = new WebClient();
        }

        #region Properties

        /// <summary>
        /// Your CryptoPhoto Application private key.
        /// </summary>
        public string PrivateKey
        { get; private set; }

        /// <summary>
        /// Your CryptoPhoto Application public key.
        /// </summary>
        public string PublicKey
        { get; private set; }

        /// <summary>
        /// The session ID provided by CryptoPhoto. This property is automatically populated if StartSession is successful.
        /// </summary>
        public string SessionID
        { get; private set; }

        #endregion

        #region StartingASession

        /// <summary>
        /// Request a new CryptoPhoto session and returns the CryptoPhoto server response as a list of strings. If successful, also sets SessionID for this object. For complete response structure and meaning refer to "https://cryptophoto.com/admin/api".
        /// </summary>
        /// <param name="uID">The unique ID associated with your user on your website.</param>
        /// <param name="ip">The IP address of the user who needs access to CryptoPhoto service.</param>
        public List<string> StartSession(int uID, string ip, bool authentication = false)
        {
            // prepare session request POST data
            NameValueCollection postData = PrepareStartSessionPostData(uID, ip);

            if (authentication) {
              postData.Add("authentication","true");
            }

            // send data to CryptoPhoto server
            string response = this.Post(START_SESSION_URL, postData);

            // translate server response
            List<string> serverResponse = response.Split('\n').ToList();

            // populate SessionID if successful
            if (serverResponse[0] == "success")
            {
                this.SessionID = serverResponse[1];
            }

            return serverResponse;
        }

        private NameValueCollection PrepareStartSessionPostData(int uID, string ip)
        {
            NameValueCollection postData = new NameValueCollection();
            // public key
            postData.Add("publickey", this.PublicKey);
            // uid
            postData.Add("uid", uID.ToString());
            // time since epoch
            long secondsSinceEpoch = (long)(DateTimeOffset.UtcNow - UnixEpoch).TotalSeconds;
            postData.Add("time", secondsSinceEpoch.ToString());
            // signature (HMAC-SHA1 of privatekey, time, uid and publickey)
            string signature = ComputeSignature(secondsSinceEpoch, uID);
            postData.Add("signature", signature);
            // ip
            postData.Add("ip", ip);

            return postData;
        }

        private string ComputeSignature(long secondsSinceEpoch, int uID)
        {
            // initialize HMACSHA1 with our private key
            byte[] privateKey = Encoding.ASCII.GetBytes(this.PrivateKey);
            HMACSHA1 hmacSha1 = new HMACSHA1(privateKey);

            // build the signature string (concatenation of privatekey, time, uid and publickey)
            string sigString = string.Format("{0}{1}{2}{3}", this.PrivateKey, secondsSinceEpoch, uID, this.PublicKey);
            byte[] sigStringBytes = Encoding.ASCII.GetBytes(sigString);

            // compute HMACSHA1 of the signature string
            byte[] sigSha1Bytes = hmacSha1.ComputeHash(sigStringBytes);
            string sigSha1String = string.Concat(Array.ConvertAll(sigSha1Bytes, b => b.ToString("X2")));

            return sigSha1String;
        }

        #endregion

        #region WidgetGeneration

        /// <summary>
        /// Generates javascript code for getting a token. Should only be called after a successful StartSession call otherwise throws an exception.
        /// </summary>
        /// <param name="sessionID">The session ID retrieved using StartSession.</param>
        public string GetGenWidget()
        {
            if (string.IsNullOrEmpty(this.SessionID))
            {
                throw new InvalidOperationException("A session must be generated using StartSession first.");
            }

            string tokenGenerationUrl = string.Format(GENERATE_WIDGET_URL_COMPOSITE, this.SessionID);
            string tokenGenerationWidgetScript = string.Format("<script type='text/javascript' src='{0}'></script>", tokenGenerationUrl);

            return tokenGenerationWidgetScript;
        }

        /// <summary>
        /// Generates javascript code for getting a challenge. Should only be called after a successful StartSession call otherwise throws an exception.
        /// </summary>
        /// <param name="sessionID">The session ID retrieved using StartSession.</param>
        public string GetAuthWidget()
        {
            if (string.IsNullOrEmpty(this.SessionID))
            {
                throw new InvalidOperationException("A session must be generated using StartSession first.");
            }

            string challengeUrl = string.Format(CHALLENGE_URL_COMPOSITE, this.SessionID);
            string challengeWidgetScript = string.Format("<script type='text/javascript' src='{0}'></script>", challengeUrl);

            return challengeWidgetScript;
        }

        #endregion

        #region ChallengeVerification

        /// <summary>
        /// post and check a user's response for a given challenge and returns the CryptoPhoto server response as a list of strings. For complete response structure and meaning refer to "https://cryptophoto.com/admin/api".
        /// </summary>
        /// <param name="selector">The value of "token_selector" sent via the login form. It corresponds to the "selector" parameter required by the CryptoPhoto verification API.</param>
        /// <param name="responseRow">The value of "token_response_field_row" sent via the login form. It corresponds to the "response_row" parameter required by the CryptoPhoto verification API.</param>
        /// <param name="responseColumn">The value of "token_response_field_col" sent via the login form. It corresponds to the "response_col" parameter required by the CryptoPhoto verification API.</param>
        /// <param name="cph">The value of "cp_phc" sent via the login form. It corresponds to the "cph" parameter required by the CryptoPhoto verification API.</param>
        /// <param name="ip">The IP address of the user who solved the CryptoPhoto Challenge.</param>
        public List<string> VerifyResponse(string selector, string responseRow, string responseColumn, string cph, int uID, string ip)
        {
            // prepare verification request POST data
            NameValueCollection postData = PrepareVerificationPostData(selector, responseRow, responseColumn, cph, uID, ip);

            // send data to CryptoPhoto server
            string response = this.Post(VERIFY_URL, postData);

            // translate server response
            List<string> serverResponse = response.Split('\n').ToList();

            return serverResponse;
        }

        private NameValueCollection PrepareVerificationPostData(string selector, string responseRow, string responseColumn, string cph, int uID, string ip)
        {
            NameValueCollection postData = new NameValueCollection();
            // public key
            postData.Add("publickey", this.PublicKey);
            // uid
            postData.Add("uid", uID.ToString());
            // time since epoch
            long secondsSinceEpoch = (long)(DateTimeOffset.UtcNow - UnixEpoch).TotalSeconds;
            postData.Add("time", secondsSinceEpoch.ToString());
            // signature (HMAC-SHA1 of privatekey, time, uid and publickey)
            string signature = ComputeSignature(secondsSinceEpoch, uID);
            postData.Add("signature", signature);
            // response row
            postData.Add("response_row", responseRow);
            // response column
            postData.Add("response_col", responseColumn);
            // selector
            postData.Add("selector", selector);
            // cph
            postData.Add("cph", cph);
            // ip
            postData.Add("ip", ip);

            return postData;
        }

        #endregion

        /// <summary>
        /// Sends a Post request with the provided data to the target URL and returns the response.
        /// </summary>
        /// <param name="address">URL to post to.</param>
        /// <param name="postData">POST data to upload.</param>
        public string Post(string url, NameValueCollection postData)
        {
            byte[] responseBytes = webClient.UploadValues(url, postData);
            string response = Encoding.UTF8.GetString(responseBytes);
            return response;
        }
    }
}
